@@ -5,12 +5,12 @@
 # pylint: disable=bare-except
 
 """
-Mie scattering calculations for perfect spheres.
+Mie scattering calculations for perfect spheres JITTED!.
 
 Extensive documentation is at <https://miepython.readthedocs.io>
 
 `miepython` is a pure Python module to calculate light scattering of
-a plane wave by non-absorbing, partially-absorbing, or perfectly conducting
+a plane wave by non-np.absorbing, partially-np.absorbing, or perfectly conducting
 spheres.
 
 The extinction efficiency, scattering efficiency, backscattering, and
@@ -39,6 +39,7 @@ Normalized Mie scattering intensities for angles mu=cos(theta)::
 """
 
 import numpy as np
+from numba import njit, int32, float64, complex128
 
 __all__ = ('ez_mie',
            'ez_intensities',
@@ -52,7 +53,7 @@ __all__ = ('ez_mie',
            'mie_mu_with_uniform_cdf',
            )
 
-
+@njit((complex128, int32), cache=True)
 def _Lentz_Dn(z, N):
     """
     Compute the logarithmic derivative of the Ricatti-Bessel function.
@@ -74,7 +75,7 @@ def _Lentz_Dn(z, N):
     ratio = alpha_j1 / alpha_j2
     runratio = alpha * ratio
 
-    while abs(abs(ratio) - 1.0) > 1e-12:
+    while np.abs(np.abs(ratio) - 1.0) > 1e-12:
         aj = zinv - aj
         alpha_j1 = 1.0 / alpha_j1 + aj
         alpha_j2 = 1.0 / alpha_j2 + aj
@@ -84,47 +85,39 @@ def _Lentz_Dn(z, N):
 
     return -N / z + runratio
 
-
-def _D_downwards(z, N):
+@njit((complex128, int32, complex128[:]), cache=True)
+def _D_downwards(z, N, D):
     """
     Compute the logarithmic derivative by downwards recurrence.
 
     Args:
         z: function argument
         N: order of Ricatti-Bessel function
-
-    Returns:
-        All the Ricatti-Bessel function values for orders from 0 to N for an
-        argument z using the downwards recurrence relations.
+        D: gets filled with the Ricatti-Bessel function values for orders
+           from 0 to N for an argument z using the downwards recurrence relations.
     """
-    D = np.zeros(N, dtype=complex)
     last_D = _Lentz_Dn(z, N)
     for n in range(N, 0, -1):
         last_D = n / z - 1.0 / (last_D + n / z)
         D[n - 1] = last_D
-    return D
 
-
-def _D_upwards(z, N):
+@njit((complex128, int32, complex128[:]), cache=True)
+def _D_upwards(z, N, D):
     """
     Compute the logarithmic derivative by upwards recurrence.
 
     Args:
         z: function argument
         N: order of Ricatti-Bessel function
-
-    Returns:
-        All the Ricatti-Bessel function values for orders from 0 to N for an
-        argument z using the upwards recurrence relations.
+        D: gets filled with the Ricatti-Bessel function values for orders
+           from 0 to N for an argument z using the upwards recurrence relations.
     """
-    D = np.zeros(N, dtype=complex)
     exp = np.exp(-2j * z)
     D[1] = -1 / z + (1 - exp) / ((1 - exp) / z - 1j * (1 + exp))
     for n in range(2, N):
         D[n] = 1 / (n / z - D[n - 1]) - n / z
-    return D
-
 
+@njit((complex128, float64, int32), cache=True)
 def _D_calc(m, x, N):
     """
     Compute the logarithmic derivative using best method.
@@ -138,15 +131,17 @@ def _D_calc(m, x, N):
         The values of the Ricatti-Bessel function for orders from 0 to N.
     """
     n = m.real
-    kappa = abs(m.imag)
+    kappa = np.abs(m.imag)
+    D = np.zeros(N, dtype=np.complex128)
 
     if n < 1 or n > 10 or kappa > 10 or x*kappa >= 3.9 - 10.8 * n + 13.78 * n**2:
-        return _D_downwards(m*x, N)
+        _D_downwards(m*x, N, D)
+    else:
+        _D_upwards(m*x, N, D)
+    return D
 
-    return _D_upwards(m*x, N)
-
-
-def _mie_An_Bn(m, x):
+@njit((complex128, float64, complex128[:], complex128[:]), cache=True)
+def _mie_An_Bn(m, x, a, b):
     """
     Compute arrays of Mie coefficients A and B for a sphere.
 
@@ -161,38 +156,37 @@ def _mie_An_Bn(m, x):
     Returns:
         An, Bn: arrays of Mie coefficents
     """
-    nstop = int(x + 4.05 * x**0.33333 + 2.0) + 1
-
-    if m.real > 0.0:
-        D = _D_calc(m, x, nstop + 1)
-
-    a = np.zeros(nstop - 1, dtype=complex)
-    b = np.zeros(nstop - 1, dtype=complex)
-
     psi_nm1 = np.sin(x)                   # nm1 = n-1 = 0
     psi_n = psi_nm1 / x - np.cos(x)       # n = 1
     xi_nm1 = complex(psi_nm1, np.cos(x))
     xi_n = complex(psi_n, np.cos(x) / x + np.sin(x))
 
-    for n in range(1, nstop):
-        if m.real == 0.0:
-            a[n - 1] = (n * psi_n / x - psi_nm1) / (n * xi_n / x - xi_nm1)
-            b[n - 1] = psi_n / xi_n
-        else:
+    nstop = len(a)
+    if m.real > 0.0:
+        D = _D_calc(m, x, nstop + 1)
+
+        for n in range(1, nstop):
             temp = D[n] / m + n / x
             a[n - 1] = (temp * psi_n - psi_nm1) / (temp * xi_n - xi_nm1)
             temp = D[n] * m + n / x
             b[n - 1] = (temp * psi_n - psi_nm1) / (temp * xi_n - xi_nm1)
+            xi = (2 * n + 1) * xi_n / x - xi_nm1
+            xi_nm1 = xi_n
+            xi_n = xi
+            psi_nm1 = psi_n
+            psi_n = xi_n.real
 
-        xi = (2 * n + 1) * xi_n / x - xi_nm1
-        xi_nm1 = xi_n
-        xi_n = xi
-        psi_nm1 = psi_n
-        psi_n = xi_n.real
-
-    return [a, b]
-
+    else:
+        for n in range(1, nstop):
+            a[n - 1] = (n * psi_n / x - psi_nm1) / (n * xi_n / x - xi_nm1)
+            b[n - 1] = psi_n / xi_n
+            xi = (2 * n + 1) * xi_n / x - xi_nm1
+            xi_nm1 = xi_n
+            xi_n = xi
+            psi_nm1 = psi_n
+            psi_n = xi_n.real
 
+@njit((complex128, float64), cache=True)
 def _small_conducting_mie(m, x):
     """
     Calculate the efficiencies for a small conducting spheres.
@@ -217,19 +211,19 @@ def _small_conducting_mie(m, x):
     ahat2 = complex(0.0, x**2 / 30.)
     bhat2 = complex(0.0, -x**2 / 45.)
 
-    qsca = x**4 * (6 * abs(ahat1)**2 + 6 * abs(bhat1)**2 + 10 * abs(ahat2)**2 +
-                   10 * abs(bhat2)**2)
+    qsca = x**4 * (6 * np.abs(ahat1)**2 + 6 * np.abs(bhat1)**2 + 10 * np.abs(ahat2)**2 +
+                   10 * np.abs(bhat2)**2)
     qext = qsca
     g = ahat1.imag * (ahat2.imag + bhat1.imag)
     g += bhat2.imag * (5.0 / 9.0 * ahat2.imag + bhat1.imag)
     g += ahat1.real * bhat1.real
     g *= 6 * x**4 / qsca
 
-    qback = 9 * x**4 * abs(ahat1 - bhat1 - 5 / 3 * (ahat2 - bhat2))**2
+    qback = 9 * x**4 * np.abs(ahat1 - bhat1 - 5 / 3 * (ahat2 - bhat2))**2
 
     return [qext, qsca, qback, g]
 
-
+@njit((complex128, float64), cache=True)
 def _small_mie(m, x):
     """
     Calculate the efficiencies for a small sphere.
@@ -257,7 +251,7 @@ def _small_mie(m, x):
     ahat2 = 1j * x2 * (m2 - 1) / 15 * (1 - x2 / 14) / \
         (2 * m2 + 3 - (2 * m2 - 7) / 14 * x2)
 
-    T = abs(ahat1)**2 + abs(bhat1)**2 + 5 / 3 * abs(ahat2)**2
+    T = np.abs(ahat1)**2 + np.abs(bhat1)**2 + 5 / 3 * np.abs(ahat2)**2
     temp = ahat2 + bhat1
     g = (ahat1 * temp.conjugate()).real / T
 
@@ -269,11 +263,11 @@ def _small_mie(m, x):
         qext = 6 * x * (ahat1 + bhat1 + 5 * ahat2 / 3).real
 
     sback = 1.5 * x**3 * (ahat1 - bhat1 - 5 * ahat2 / 3)
-    qback = 4*abs(sback)**2/x2
+    qback = 4*np.abs(sback)**2/x2
 
     return [qext, qsca, qback, g]
 
-
+@njit((complex128, float64), cache=True)
 def _mie_scalar(m, x):
     """
     Calculate the efficiencies for a sphere when both m and x are scalars.
@@ -289,37 +283,40 @@ def _mie_scalar(m, x):
         g: the average cosine of the scattering phase function
     """
     if m.real == 0 and x < 0.1:
-        return _small_conducting_mie(m, x)
+        qext, qsca, qback, g =_small_conducting_mie(m, x)
 
-    if m.real > 0.0 and abs(m) * x < 0.1:
-        return _small_mie(m, x)
+    elif m.real > 0.0 and np.abs(m) * x < 0.1:
+        qext, qsca, qback, g =_small_mie(m, x)
 
-    a, b = _mie_An_Bn(m, x)
+    else:
+        nstop = int(x + 4.05 * x**0.33333 + 2.0) + 1
+        a = np.zeros(nstop - 1, dtype=np.complex128)
+        b = np.zeros(nstop - 1, dtype=np.complex128)
+        _mie_An_Bn(m, x, a, b)
 
-    nmax = len(a)
-    n = np.arange(1, nmax + 1)
-    cn = 2.0 * n + 1.0
+        nmax = len(a)
+        n = np.arange(1, nmax + 1)
+        cn = 2.0 * n + 1.0
 
-    qext = 2 * np.sum(cn * (a.real + b.real)) / x**2
-    qsca = qext
+        qext = 2 * np.sum(cn * (a.real + b.real)) / x**2
+        qsca = qext
 
-    if m.imag != 0:
-        qsca = 2 * np.sum(cn * (abs(a)**2 + abs(b)**2)) / x**2
+        if m.imag != 0:
+            qsca = 2 * np.sum(cn * (np.abs(a)**2 + np.abs(b)**2)) / x**2
 
-    qback = abs(np.sum((-1)**n * cn * (a - b)))**2 / x**2
+        qback = np.abs(np.sum((-1)**n * cn * (a - b)))**2 / x**2
 
-    c1n = n * (n + 2) / (n + 1)
-    c2n = cn / n / (n + 1)
-    g = 0
-    for i in range(nmax - 1):
-        asy1 = c1n[i] * (a[i] * a[i + 1].conjugate() +
-                         b[i] * b[i + 1].conjugate()).real
-        asy2 = c2n[i] * (a[i] * b[i].conjugate()).real
-        g += 4 * (asy1 + asy2) / qsca / x**2
+        c1n = n * (n + 2) / (n + 1)
+        c2n = cn / n / (n + 1)
+        g = 0
+        for i in range(nmax - 1):
+            asy1 = c1n[i] * (a[i] * a[i + 1].conjugate() +
+                             b[i] * b[i + 1].conjugate()).real
+            asy2 = c2n[i] * (a[i] * b[i].conjugate()).real
+            g += 4 * (asy1 + asy2) / qsca / x**2
 
     return [qext, qsca, qback, g]
 
-
 def mie(m, x):
     """
     Calculate the efficiencies for a sphere where m or x may be arrays.
@@ -334,29 +331,30 @@ def mie(m, x):
         qback: the backscatter efficiency
         g: the average cosine of the scattering phase function
     """
-    try:
-        mlen = len(m)
-    except:
-        mlen = 0
-        mm = m
+    mm = m
+    xx = x
 
-    try:
-        xlen = len(x)
-    except:
-        xlen = 0
-        xx = x
-
-    if xlen == 0 and mlen == 0:
+    if np.isscalar(m) and np.isscalar(x):
         return _mie_scalar(mm, xx)
 
-    if xlen and mlen and xlen != mlen:
+    if np.isscalar(m):
+        mlen = 0
+    else:
+        mlen = len(m)
+
+    if np.isscalar(x):
+        xlen = 0
+    else:
+        xlen = len(x)
+
+    if xlen > 0 and mlen > 0 and xlen != mlen:
         raise RuntimeError('m and x arrays to mie must be same length')
 
     thelen = max(xlen, mlen)
-    qext = np.empty(thelen)
-    qsca = np.empty(thelen)
-    qback = np.empty(thelen)
-    g = np.empty(thelen)
+    qext = np.empty(thelen, dtype=np.float64)
+    qsca = np.empty(thelen, dtype=np.float64)
+    qback = np.empty(thelen, dtype=np.float64)
+    g = np.empty(thelen, dtype=np.float64)
 
     for i in range(thelen):
         if mlen > 0:
@@ -369,7 +367,7 @@ def mie(m, x):
 
     return qext, qsca, qback, g
 
-
+@njit((complex128, float64, float64[:]), cache=True)
 def _small_mie_conducting_S1_S2(m, x, mu):
     """
     Calculate the scattering amplitudes for small conducting spheres.
@@ -398,15 +396,15 @@ def _small_mie_conducting_S1_S2(m, x, mu):
     S2 = 1.5 * x**3 * (bhat1 + ahat1 * mu + 5 / 3 * bhat2 *
                        mu + 5 / 3 * ahat2 * (2 * mu**2 - 1))
 
-    qext = x**4 * (6 * abs(ahat1)**2 + 6 * abs(bhat1)**2 + 10 * abs(ahat2)**2 +
-                   10 * abs(bhat2)**2)
+    qext = x**4 * (6 * np.abs(ahat1)**2 + 6 * np.abs(bhat1)**2 + 10 * np.abs(ahat2)**2 +
+                   10 * np.abs(bhat2)**2)
     norm = np.sqrt(qext * np.pi * x**2)
     S1 /= norm
     S2 /= norm
 
     return [S1, S2]
 
-
+@njit((complex128, float64, float64[:]), cache=True)
 def _small_mie_S1_S2(m, x, mu):
     """
     Calculate the scattering amplitude functions for small spheres (x<0.1).
@@ -448,7 +446,7 @@ def _small_mie_S1_S2(m, x, mu):
 
     return [S1, S2]
 
-
+@njit((complex128, float64, float64[:]), cache=True)
 def mie_S1_S2(m, x, mu):
     """
     Calculate the scattering amplitude functions for spheres.
@@ -466,11 +464,14 @@ def mie_S1_S2(m, x, mu):
     Returns:
         S1, S2: the scattering amplitudes at each angle mu [sr**(-0.5)]
     """
-    a, b = _mie_An_Bn(m, x)
+    nstop = int(x + 4.05 * x**0.33333 + 2.0) + 1
+    a = np.zeros(nstop - 1, dtype=np.complex128)
+    b = np.zeros(nstop - 1, dtype=np.complex128)
+    _mie_An_Bn(m, x, a, b)
 
     nangles = len(mu)
-    S1 = np.zeros(nangles, dtype=complex)
-    S2 = np.zeros(nangles, dtype=complex)
+    S1 = np.zeros(nangles, dtype=np.complex128)
+    S2 = np.zeros(nangles, dtype=np.complex128)
 
     nstop = len(a)
     for k in range(nangles):
@@ -496,7 +497,6 @@ def mie_S1_S2(m, x, mu):
 
     return [S1, S2]
 
-
 def mie_cdf(m, x, num):
     """
     Create a CDF for unpolarized scattering uniformly spaced in cos(theta).
@@ -522,7 +522,7 @@ def mie_cdf(m, x, num):
     mu = np.linspace(-1, 1, num)
     s1, s2 = mie_S1_S2(m, x, mu)
 
-    s = (abs(s1)**2 + abs(s2)**2) / 2
+    s = (np.abs(s1)**2 + np.abs(s2)**2) / 2
 
     cdf = np.zeros(num)
     total = 0
@@ -533,7 +533,6 @@ def mie_cdf(m, x, num):
 
     return mu, cdf
 
-
 def mie_mu_with_uniform_cdf(m, x, num):
     """
     Create a CDF for unpolarized scattering for uniform CDF.
@@ -589,7 +588,6 @@ def mie_mu_with_uniform_cdf(m, x, num):
 
     return [mu, cdf]
 
-
 def generate_mie_costheta(mu_cdf):
     """
     Generate a new scattering angle using a cdf.
@@ -617,7 +615,6 @@ def generate_mie_costheta(mu_cdf):
 
     return x
 
-
 def i_per(m, x, mu):
     """
     Return the scattered intensity in a plane normal to the incident light.
@@ -636,10 +633,9 @@ def i_per(m, x, mu):
        The intensity at each angle in the array mu.  Units [1/sr]
     """
     s1, _ = mie_S1_S2(m, x, mu)
-    intensity = abs(s1)**2
+    intensity = np.abs(s1)**2
     return intensity.astype('float')
 
-
 def i_par(m, x, mu):
     """
     Return the scattered intensity in a plane parallel to the incident light.
@@ -658,10 +654,9 @@ def i_par(m, x, mu):
        The intensity at each angle in the array mu.  Units [1/sr]
     """
     _, s2 = mie_S1_S2(m, x, mu)
-    intensity = abs(s2)**2
+    intensity = np.abs(s2)**2
     return intensity.astype('float')
 
-
 def i_unpolarized(m, x, mu):
     """
     Return the unpolarized scattered intensity at specified angles.
@@ -680,10 +675,9 @@ def i_unpolarized(m, x, mu):
        The intensity at each angle in the array mu.  Units [1/sr]
     """
     s1, s2 = mie_S1_S2(m, x, mu)
-    intensity = (abs(s1)**2 + abs(s2)**2) / 2
+    intensity = (np.abs(s1)**2 + np.abs(s2)**2) / 2
     return intensity.astype('float')
 
-
 def ez_mie(m, d, lambda0, n_env=1.0):
     """
     Calculate the efficiencies of a sphere.
@@ -704,7 +698,6 @@ def ez_mie(m, d, lambda0, n_env=1.0):
     x_env = np.pi*d/(lambda0/n_env)
     return mie(m_env, x_env)
 
-
 def ez_intensities(m, d, lambda0, mu, n_env=1.0):
     """
     Return the scattered intensities from a sphere.
@@ -732,8 +725,8 @@ def ez_intensities(m, d, lambda0, mu, n_env=1.0):
     lambda_env = lambda0 / n_env
     x_env = np.pi*d/lambda_env
     s1, s2 = mie_S1_S2(m_env, x_env, mu)
-    ipar = abs(s2)**2
-    iper = abs(s1)**2
+    ipar = np.abs(s2)**2
+    iper = np.abs(s1)**2
     Ipar = ipar.astype('float')
     Iper = iper.astype('float')
     return Ipar, Iper
